#!/usr/bin/env python

from cPickle import dumps, loads
from elementtree.ElementTree import Element, XML, tostring
from getopt import getopt, GetoptError
from os import close,fork
from signal import signal, SIGINT
from sys import argv,exc_info,exit
from syslog import syslog, LOG_INFO, LOG_ERR
from time import sleep
from traceback import extract_tb

from sss.restriction import Data, DataSet
from sss.server import Server, AddEvent, BufferedPipe
from sss.ssslib import comm_lib, InitError, ConnectError, CommunicationError

class Location(Data):
    def __init__(self,element):
        self.element = element
        for field in ['tier','schema_version']:
            if not self.element.attrib.get(field, None):
                self.element.attrib[field] = 1

class service_directory(Server):
    __implementation__='sd'
    __component__='service-directory'
    __statefields__=['locations']
    __dispatch__ = {'add-location':'AddLocation',
                    'get-location':'locations.Get',
                    'del-location':'locations.Del'}

    def __setup__(self):
        self.locations = DataSet("locations", "location", Location, None, False)
        (port,host,proto,sversion,tier)=map(lambda x:comm_lib().GetConfig().get(x),['sd_port','sd_host','sd_protocol','sd_schema_version','sd_tier'])
        if filter(lambda x:x == None,(port,host,proto,sversion,tier)):
            print "Insufficient data in sss.conf.  Exiting..."
            exit(1)
        myloc=XML('''<add-location><location component='service-directory' host='%s' port='%s' protocol='%s' schema_version='%s' tier='%s'/></add-location>'''%(host,port,proto,sversion,tier))
        self.locations.Add(myloc,('localhost',0))
        self.port=int(port)
        self.protocol=proto

    def AddLocation(self,xml,(peer,port)):
        return self.locations.Add(xml,(peer,port), self.SendLocEvt)

    def SendLocEvt(self, loc, arg):
        if loc.element.attrib['component'] != 'event-manager':
            self.regevt.WriteMessage(dumps(("service-directory","service-registered",loc.element.attrib['component'])))
        
    def ChildSigHand(self,signum,frame):
        exit(0)

    def __setupChildren__(self):
        self.regevt = BufferedPipe()
        self.pipes.append(self.regevt)
        pid=fork()
        if pid == 0:
            signal(SIGINT, self.ChildSigHand)
            while 1:
                msg = self.regevt.ReadMessage()
                try:
                    apply(AddEvent,loads(msg))
                except:
                    syslog(LOG_ERR, "Failed to send registration event")
        else:
            pinger_pid = fork()
            if pinger_pid == 0:
                self.RunPinger()
            else:
                close(self.regevt.r)
                return [pid,pinger_pid]

    def RunPinger(self):
        try:
            c = comm_lib()
            while 1:
                sleep(300) # 5 minutes
                h = c.ClientInit("service-directory")
                c.SendMessage(h,"<get-location><location component='*' host='*' port='*' protocol='*'/></get-location>")
                locs = XML(c.RecvMessage(h)).findall('location')
                c.ClientClose(h)
                for loc in locs:
                    cmp = loc.attrib["component"]
                    try:
                        h = c.ClientInit(cmp)
                        c.ClientClose(h)
                    except (ConnectError,CommunicationError):
                        h = c.ClientInit("service-directory")
                        c.SendMessage(h,"<del-location><location component='%s'/></del-location>"%cmp)
                        syslog(LOG_INFO,"component %s flushed from sd due to connect error"%cmp)
                        c.RecvMessage(h)
                        c.ClientClose(h)
                    except:
                        (t,v,tb)=exc_info()
                        syslog(LOG_ERR, "Failure in sd-pinger")
                        syslog(LOG_ERR, "Attempted to contact %s"%cmp)
                        for line in extract_tb(tb):
                            errstr = '  File "%s", line %i, in %s\n    %s\n'%line
                            syslog(LOG_ERR,errstr)
                            errstr = "%s: %s\n"%(t,v)
                            syslog(LOG_ERR,errstr)
                        del t,v,tb
        except KeyboardInterrupt:
            exit(1)
        except:
            (t,v,tb)=exc_info()
            syslog(LOG_ERR, "*** unexpected Failure in sd-pinger")
            for line in extract_tb(tb):
                errstr = '  File "%s", line %i, in %s\n    %s\n'%line
                syslog(LOG_ERR,errstr)
            errstr = "%s: %s\n"%(t,v)
            syslog(LOG_ERR,errstr)
            del t,v,tb
            exit(1)

if __name__ == "__main__":
    try:
        (opts,args)=getopt(argv[1:],'d',['daemon='])
    except GetoptError, msg:
        print "%s\nUsage:\nsd.py [-d] [--daemon <pidfile>]"%(msg)
        exit(1)
    daemon=filter(lambda x:(x[0] == '--daemon'),opts)
    ldebug=len(filter(lambda x:(x[0] == '-d'),opts))
    if daemon:
        from sss.daemonize import daemonize
        daemonize(daemon[0][1])
    try:
        s=service_directory(debug=ldebug)
    except InitError, msg:
        print "Library init Error:%s"%(msg)
        exit(1)
    s.ServeForever()
